from wormhole.streamer import SocketIOStreamerBase

import cv2
from typing import Optional, Any


class RawStreamer(SocketIOStreamerBase):
    """
    Raw Image Streaming
    This is generally not recommended as it sends A LOT of information over the network, 
    But its an interesting experiment to see the maximum processing throughput
    Without any image encoding
    """

    def __init__(
        self,
        *args,
        **kwargs
    ):
        # Initiate Parent SocketIO Streamer Object
        super().__init__(self.stream_hotloop, *args, **kwargs)

    # Hotloop for sending raw video
    def stream_hotloop(self):
        self.send_data(self.video.get_frame().tobytes())


class RawIMEncodeStreamerBase(SocketIOStreamerBase):
    """
    Streamer for all types of image formats supported by imencode  
    """

    def __init__(
        self,
        file_format: str,
        *args,
        imencode_config: Optional[list[Any]] = None,
        **kwargs
    ):
        # Initiate Parent SocketIO Streamer Object
        super().__init__(self.stream_hotloop, *args, **kwargs)

        # Setup some configs for the imencode function
        self.file_format = file_format
        self.imencode_config = imencode_config

    # Hotloop for sending raw video
    def stream_hotloop(self):
        _, encoded_image = cv2.imencode(
            self.file_format,
            self.video.get_frame(),
            self.imencode_config or []  # Use Empty Config if imencode is not available
        )
        self.send_data(encoded_image.tobytes())

# Proxy classes for each of the supported streaming formats.
# They all run the exact same thing, but this is here so it fits with the API structure


class RawJPEGStreamer(RawIMEncodeStreamerBase):
    """
    Raw JPEG Streaming
    Sends raw individual PNG frames over the network
    """

    def __init__(
        self,
        *args,
        **kwargs
    ):
        super().__init__(".jpeg", *args, **kwargs)


class RawPNGStreamer(RawIMEncodeStreamerBase):
    """
    Raw PNG Streaming
    Sends raw individual PNG frames over the network
    """

    def __init__(
        self,
        *args,
        **kwargs
    ):
        super().__init__(".png", *args, **kwargs)


class RawImageFormatStreamer(RawIMEncodeStreamerBase):
    """
    Raw Image File Streaming
    Sends raw individual image file frames over the network
    """

    def __init__(
        self,
        image_format,
        *args,
        **kwargs
    ):
        super().__init__(image_format, *args, **kwargs)



from wormhole.viewer import SocketIOViewerBase

import cv2
import numpy as np


class RawViewer(SocketIOViewerBase):
    """
    Viewer for Raw Image Streaming
    """

    def __init__(
        self,
        *args,
        **kwargs
    ):
        # Initiate Parent SocketIO Viewer Object
        super().__init__(self.raw_image_handler, *args, **kwargs)

    # Create Handler for Incoming Raw Data Frames
    def raw_image_handler(self, raw):
        try:
            # Sanity Check for Stream Info
            if len(raw) != self.width * self.height * self.pixel_size:
                raise Exception(f"Invalid frame size! Expected: {self.width * self.height * self.pixel_size} bytes but received: {raw} bytes!")

            # Convert 1d data array to 3d frame data
            new_frame = np.ndarray((self.width, self.height, self.pixel_size), np.uint8, raw)

            # New Frame!
            self.set_frame(new_frame)
        except Exception as e:
            self.handle_render_error(e, message="Error While Reading/Processing RAW stream!")


class RawIMDecodeViewerBase(SocketIOViewerBase):
    """
    Viewer for all types of image formats supported by imdecode
    """

    def __init__(
        self,
        *args,
        **kwargs
    ):
        # Initiate Parent SocketIO Viewer Object
        super().__init__(self.raw_image_handler, *args, **kwargs)

    # Create Handler for Incoming Raw Data Frames
    def raw_image_handler(self, raw):
        try:
            # Read and decode data
            np_image = np.fromstring(raw, dtype=np.uint8)  # type: ignore
            new_frame = cv2.imdecode(np_image, cv2.IMREAD_COLOR)

            # If sizes does not match, resize frame
            frame_height, frame_width, _ = new_frame.shape
            if frame_width != self.width or frame_height != self.height:
                new_frame = cv2.resize(new_frame, (self.width, self.height))

            # Set the new frame
            self.set_frame(new_frame)
        except Exception as e:
            self.handle_render_error(e, message="Error While Reading/Processing raw image stream!")

# Proxy classes for each of the supported streaming formats.
# They all run the exact same thing, but this is here so it fits with the API structure


class RawJPEGViewer(RawIMDecodeViewerBase):
    """
    Viewer for Raw JPEG Streaming
    """

    def __init__(
        self,
        *args,
        **kwargs
    ):
        # Initiate Parent SocketIO Viewer Object
        super().__init__(*args, **kwargs)


class RawPNGViewer(RawIMDecodeViewerBase):
    """
    Viewer for Raw PNG Streaming
    """

    def __init__(
        self,
        *args,
        **kwargs
    ):
        # Initiate Parent SocketIO Viewer Object
        super().__init__(*args, **kwargs)


class RawImageFormatViewer(RawIMDecodeViewerBase):
    """
    Viewer for Raw Image Format Streaming
    """

    def __init__(
        self,
        image_format,
        *args,
        **kwargs
    ):
        # Image Format isnt used, but keeping it here for API compatibility.
        self.image_format = image_format
        # Initiate Parent SocketIO Viewer Object
        super().__init__(*args, **kwargs)



from wormhole.controller import AbstractController
from wormhole.video import AbstractVideo

from typing import Optional


class AbstractStreamer():
    """
    General Abstract Streamer Class for Wormhole.
    """

    def __init__(
        self,
        controller: AbstractController,
        video: AbstractVideo,
        route: str,
        fps_override: Optional[float] = None,
        print_fps: bool = False,
        strict_url: bool = True
    ):
        self.controller = controller
        self.video = video
        self.route = route
        self.max_fps = fps_override or self.video.max_fps
        self.print_fps = print_fps
        self.strict_url = strict_url



import logging
import numpy as np
import time
import traceback
from typing import Callable
from wormhole.utils import blank_frame_color, draw_text, FrameController


class AbstractVideo():
    """
    General Abstract Video Class for Wormhole. Has some advanced features that are helpful for video processing
    """

    def __init__(
        self,
        width: int,
        height: int,
        max_fps: float,
        pixel_size: int = 3,
        print_fps: bool = False,
        frame_modifiers=None,
        frame_subscribers=None
    ):
        # Basic Video Properties
        self.width: int = width
        self.pixel_size: int = pixel_size
        self.height: int = height
        self.max_fps: float = max_fps
        self.print_fps: bool = print_fps

        # Sanity Check
        if 0 > self.width or 0 > self.height or 0 > self.max_fps:
            raise ValueError("Video Properties cannot smaller than 0!")

        # Video Information
        self._frame: np.ndarray = np.zeros((width, height, self.pixel_size), np.uint8)
        self.finished_frame: np.ndarray = self._frame

        # Frame Modifiers -> List of functions that change the output video when a new frame arrives
        self.frame_modifiers: list[Callable[[AbstractVideo], None]] = frame_modifiers or []
        # Frame Subscribers -> List of functions to call when a new frame arrives
        self.frame_subscribers: list[Callable[[AbstractVideo], None]] = frame_subscribers or []
        # Set up Frame Controller
        self.frame_controller = FrameController(self.max_fps, print_fps=self.print_fps)

    # Add a function to the frame modifiers
    def add_frame_modifier(self, modifier):
        self.frame_modifiers.append(modifier)

    # Add a function to the frame subscribers
    def add_frame_subscriber(self, subscriber):
        self.frame_subscribers.append(subscriber)

    # Call all frame modifiers
    def call_frame_modifiers(self):
        for modifier in self.frame_modifiers:
            try:
                modifier(self)
            except Exception as error:
                logging.error(f"Error While Running Frame Modifier {modifier}")
                traceback.print_exc()

                # Render error to video frame
                draw_text(self._frame, "ERROR!", (10, 60), font_color=(0, 0, 255), font_size=2, font_stroke=4)
                draw_text(self._frame, f"Error While Running Frame Modifier {modifier}!", (10, 100))
                draw_text(self._frame, f"Error: {error}", (10, 130), font_size=0.5, font_stroke=1)

    # Call all frame subscribers
    def call_frame_subscribers(self):
        for subscriber in self.frame_subscribers:
            try:
                subscriber(self)
            except Exception as e:
                logging.error(f"Error While Running Frame Subscriber {subscriber}")
                traceback.print_exc()

    # Get the current frame
    def get_frame(self):
        return self.finished_frame

    # Set the current frame
    def set_frame(self, frame: np.ndarray):
        # Sanity Check Frame Size
        if frame.size != self.width * self.height * self.pixel_size:
            raise ValueError(f"Frame Size Does Not Match! Frame Size: {frame.size}, Expected Size: {self.height * self.width * self.pixel_size}")

        # Set Frame
        self._frame = frame
        self.call_frame_modifiers()
        self.finished_frame = self._frame
        self.call_frame_subscribers()

    # Set the current frame to a blank frame
    def set_blank_frame(self):
        new_frame = np.zeros((self.width, self.height, self.pixel_size), np.uint8)
        self.set_frame(new_frame)

    def handle_render_error(self, error, message="Error While Generating Next Frame!"):
        try:
            # Print error to user
            logging.error(f"Error While Rendering Frame: {error}")
            traceback.print_exc()

            # Render an error video frame
            error_frame = blank_frame_color(self.width, self.height, (0, 0, 0))
            error_frame = draw_text(error_frame, "ERROR!", (10, 60), font_color=(0, 0, 255), font_size=2, font_stroke=4)
            error_frame = draw_text(error_frame, message, (10, 100))
            error_frame = draw_text(error_frame, f"Error: {error}", (10, 130), font_size=0.5, font_stroke=1)
            self.finished_frame = error_frame

            # Sleep one second so its not hotlooping like crazy
            time.sleep(1)

            # Reset FPS statistics in case the video works again
            self.frame_controller.reset_fps_stats()
        except Exception as e:
            print(f"Error while processing error frame!!!!! {e}")
            print(f"Something is seriously wrong with this video object or this instance of Wormhole!")



from wormhole.video.abstractvideo import *
from wormhole.video.customvideo import *
from wormhole.video.cameravideo import *
from wormhole.video.filevideo import *
from wormhole.video.imagevideo import *
from wormhole.video.videocopy import *
from wormhole.video.videorender import *
from wormhole.video.videowriter import *



import logging
import requests
import threading
import traceback
from flask import request
from markupsafe import escape
from threading import Thread
from typing import Optional, Type

from wormhole.version import __version__
from wormhole.controller import AbstractController, FlaskController
from wormhole.streamer import AbstractStreamer
from wormhole.viewer import AbstractViewer
from wormhole.video import AbstractVideo


class Wormhole():
    """
    Entrypoint for everything WORMHOLE!
    """

    #
    # --- Basic Wormhole Setup ---
    #

    def __init__(
        self,
        # Basic Controller Settings
        network_controller: Type[AbstractController] = FlaskController,
        # Advanced Controller Settings
        # Extra Features
        welcome_screen: bool = True,
        advanced_features: bool = True,
        supported_protocols: Optional[
            dict[
                str,
                tuple[
                    Type[AbstractStreamer],
                    Type[AbstractViewer]
                ]
            ]
        ] = None,
        debug: bool = False,
        # Pass All Other Arguments to the Controller
        *args,
        **kwargs
    ):
        # Set logging if debug
        if debug:
            logging.basicConfig(level=logging.DEBUG)
        logging.info(f"Setting Up Wormhole Server {__version__}")

        # Create Controller
        self.controller = network_controller(self, *args, debug=debug, **kwargs)

        # Set Up Streamers
        # List of streamers with the route as they key and the streamer object as the value
        self.routes: dict[str, AbstractStreamer] = {}
        # List of automatically managed streams with the video name as the key
        # The value is a tuple with the stream object and the list of supported protocols
        self.managed_streams: dict[str, tuple[AbstractVideo, list[str]]] = {}

        # Set up advanced Wormhole features
        self.advanced_features = advanced_features
        from wormhole.streamer import (
            RawJPEGStreamer,
            MJPEGStreamer
        )
        from wormhole.viewer import (
            RawJPEGViewer,
            MJPEGViewer
        )
        self.supported_protocols = supported_protocols or {
            # ORDER MATTERS HERE! Ranked in order from most preferred to least preferred!
            "RAWJPEG": (RawJPEGStreamer, RawJPEGViewer),
            "MJPEG": (MJPEGStreamer, MJPEGViewer),
        }
        if self.advanced_features:
            if len(self.supported_protocols) == 0:
                raise Exception("No supported protocols were passed!")
            self.set_up_advanced_features()
        logging.debug(f"Supported Protocols: {list(self.supported_protocols.keys())}")

        # Set up welcome screen
        if welcome_screen:
            def hello_world():
                return (
                    "<h1>Welcome to Wormhole!</h1>"
                    "<h3>If you see this page, Wormhole Video Streaming Server is successfully installed and working!</h3>"
                    f"<p>Version: {__version__}</p>"
                    f"{self.generate_debug_html() if self.controller.debug else ''}"  # type: ignore
                )
            self.controller.add_route("/", hello_world, strict_slashes=False)

        # Start Server In Another Thread
        self.wormhole_thread = Thread(target=self.controller.start_server, daemon=True)
        self.wormhole_thread.start()

    #
    # --- Advanced Wormhole Setup ---
    #

    def set_up_advanced_features(self):
        logging.debug("Setting up advanced Wormhole features")

        # Set up basic client sync
        def client_sync():
            # Default Response
            server_response = {
                "version": __version__,
                "supported_protocols": list(self.supported_protocols.keys()),
                "managed_streams": list(self.managed_streams.keys()),
            }

            # Get client information
            if not request.json:
                return {
                    "ready": False,
                    "message": "Invalid Client Information Sent!",
                    **server_response
                }, 400
            if not all(key in request.json for key in ["version", "supported_protocols"]):
                return {
                    "ready": False,
                    "message": "Posted Json Is Missing Fields!",
                    **server_response
                }, 400

            # Check if versions match
            client_version = request.json.get("version", "N/A")
            if client_version != __version__:
                return {
                    "ready": False,
                    "message": f"Wormhole Version Mismatch! Server: {__version__} Client: {client_version}",
                    **server_response
                }, 400

            # Get client supported protocols & verify if they are supported
            client_supported_protocols = request.json.get("supported_protocols", [])
            if not any([p in client_supported_protocols for p in self.supported_protocols.keys()]):
                return {
                    "ready": False,
                    "message": f"Client Does Not Support Any of The Supported Protocols! Server Supports: {list(self.supported_protocols.keys())} Client Supports: {client_supported_protocols}",
                    **server_response
                }, 400

            # Return with server information
            return {
                "ready": True,
                "message": f"Wormhole Is Ready To Connect!",
                **server_response
            }, 200

        self.controller.add_route("/wormhole/sync", client_sync, methods=["POST"], strict_slashes=False, strict_url=False)

        # Set Up Stream Sync
        def stream_sync(name):
            if name not in self.managed_streams:
                return "Stream Not Found!", 404

            # Get Stream Information
            video_obj, supported_protocols = self.managed_streams[name]

            # Return with server information
            server_response = {
                "version": __version__,
                "stream_name": name,
                "supported_protocols": supported_protocols,
                "stream_info": {
                    "width": video_obj.width,
                    "height": video_obj.height,
                    "pixel_size": video_obj.pixel_size,
                    "max_fps": video_obj.max_fps
                }
            }

            return server_response, 200
        self.controller.add_route("/wormhole/stream/<name>/sync", stream_sync, methods=["GET", "POST"], strict_slashes=False, strict_url=False)

    #
    # --- Managed Wormhole Streaming ---
    #

    # Simple Aliases For The Different Video Types
    def stream(self, resource: str, *args, **kwargs):
        # Check if stream location is a camera
        if str(resource).isnumeric():
            return self.stream_camera(int(resource), *args, **kwargs)
        # Else, assume its a file
        else:
            return self.stream_file(resource, *args, **kwargs)

    def stream_camera(
        self,
        cam_id: int,
        name: str = "default",
        protocols: Optional[list[str]] = None,
        camera_fps: Optional[int] = None,
        stream_fps: Optional[int] = None,
        print_camera_fps: bool = False,
        print_stream_fps: bool = False,
        **kwargs
    ):
        # Separate out kwargs for CameraVideo object or Streamer Object
        # As CameraVideo is constant, we can hardcode these.
        camera_video_arg_keys = ["width", "height", "cv2_config", "pixel_size", "frame_modifiers", "frame_subscribers"]
        camera_video_args = {}
        streamer_args = {}
        for key, value in kwargs.items():
            if key in camera_video_arg_keys:
                camera_video_args[key] = value
            else:
                streamer_args[key] = value

        # Initialize Camera Video
        from wormhole.video import CameraVideo
        video = CameraVideo(cam_id, max_fps=camera_fps, print_fps=print_camera_fps, **camera_video_args)

        logging.info("Creating Managed Video Stream From Camera!")
        self.stream_video(video, name=name, protocols=protocols, fps_override=stream_fps, print_fps=print_stream_fps, **streamer_args)

    def stream_file(
        self,
        filename: str,
        name: str = "default",
        protocols: Optional[list[str]] = None,
        video_fps: Optional[int] = None,
        stream_fps: Optional[int] = None,
        print_video_fps: bool = False,
        print_stream_fps: bool = False,
        **kwargs
    ):
        # Separate out kwargs for FileVideo object or Streamer Object
        # As FileVideo is constant, we can hardcode these.
        file_video_arg_keys = ["width", "height", "repeat", "cv2_config", "pixel_size", "frame_modifiers", "frame_subscribers"]
        file_video_args = {}
        streamer_args = {}
        for key, value in kwargs.items():
            if key in file_video_arg_keys:
                file_video_args[key] = value
            else:
                streamer_args[key] = value

        # Initialize File Video
        from wormhole.video import FileVideo
        video = FileVideo(filename, max_fps=video_fps, print_fps=print_video_fps, **file_video_args)

        logging.info("Creating Managed Video Stream From File!")
        self.stream_video(video, name=name, protocols=protocols, fps_override=stream_fps, print_fps=print_stream_fps, **streamer_args)

    def stream_video(self, video: AbstractVideo, *args, name: str = "default", protocols: Optional[list[str]] = None, **kwargs):
        # Check if advanced features are enabled
        if not self.advanced_features:
            raise Exception("Managed Streams Are Only Enabled If Advanced Features Are Enabled!")

        # Fill in list of protocols if not passed in
        protocols = protocols or list(self.supported_protocols.keys())
        if len(protocols) == 0:
            raise Exception("Protocols List Is Empty!")
        if any([p not in self.supported_protocols.keys() for p in protocols]):
            raise Exception("Protocols List Contains Unsupported Protocols!")
        logging.debug(f"Wormhole Sync: Server Supports {protocols}")

        # Process Name
        name = name.lower()
        if not name.isalnum():
            raise Exception("Name Must Be Alphanumeric!")
        if name in self.managed_streams:
            raise Exception(f"Name {name} Is Already Used!")
        logging.debug(f"Wormhole Sync: Creating Managed Stream With Name: {name}")

        # For each given protocol, start streaming!
        for proto in protocols:
            # Get the streamer class
            streamer, _ = self.supported_protocols[proto]
            # Initialize the streamer
            logging.debug(f"Wormhole Sync: Setting Up Stream Process With Protocol {proto}")
            self.create_stream(streamer, video, f"/wormhole/stream/{name}/{proto.lower()}", *args, strict_url=False, **kwargs)

        # Add the streamer to the list of managed streams
        self.managed_streams[name] = (video, protocols)
        return video

    #
    # --- Managed Wormhole Viewing ---
    #

    def view(self, hostname: str, name: str = "default"):
        # Check if advanced features are enabled
        if not self.advanced_features:
            raise Exception("Managed Streams Are Only Enabled If Advanced Features Are Enabled!")

        # Process Name
        name = name.lower()
        if not name.isalnum():
            raise Exception("Name Must Be Alphanumeric!")
        logging.debug(f"Wormhole Sync: Viewing Managed Stream With Name: {name}")

        # Verify hostname is valid
        if not hostname.startswith("http://") and not hostname.startswith("https://"):
            hostname = f"http://{hostname}"
        if hostname.endswith("/"):
            hostname = hostname[:-1]
        logging.debug(f"Wormhole Sync: Connecting To Host: {hostname}")
        # TODO: use regex to validate hostname, but idk regex

        # Sync with the server for information
        server_streams = self.sync_wormhole(hostname)
        if name not in server_streams:
            raise Exception(f"Requested Stream {name} not being streamed by the server! The current streams are: {server_streams}")
        logging.debug(f"Wormhole Sync: Stream {name} is available on the server!")

        # Sync Stream Information
        stream_protocols, stream_width, stream_height, stream_pixel_size, stream_fps = self.sync_stream(hostname, name)
        logging.debug(f"Wormhole Sync: Stream {name} supports protocols: {stream_protocols} and has args: Width: {stream_width}, Height: {stream_height}, Pixel Size: {stream_pixel_size}, FPS: {stream_fps}!")

        # Do an intersection between stream_protocols and self.supported_protocols
        # to get the list of protocols we can use
        common_protocols = [proto for proto in stream_protocols if proto in self.supported_protocols]
        logging.debug(f"Wormhole Sync: Common protocols are: {common_protocols} that we can use!")
        if len(common_protocols) == 0:
            raise Exception(f"None of the protocols supported by the server are supported by this client! Server: {stream_protocols}, Client: {self.supported_protocols}")

        # Try each supported protocol
        for proto in common_protocols:
            logging.debug(f"Attempting to view stream with protocol {proto}")

            try:
                # Get the viewer class
                _, viewer = self.supported_protocols[proto]

                # Initialize the viewer
                logging.debug(f"Wormhole Sync: Attempting to Initializing Viewer with {proto}!")
                viewer_obj = viewer(f"{hostname}/wormhole/stream/{name}/{proto.lower()}", stream_width, stream_height, max_fps=stream_fps, pixel_size=stream_pixel_size)

                logging.debug(f"Success! Using {proto} for streaming")
                return viewer_obj
            except Exception as e:
                logging.error(f"Failed to initialize stream with protocol {proto}! Error: {e}")
                traceback.print_exc()
                logging.error(f"Attempting Fallback!")
        else:
            raise Exception(f"No Supported Protocols Was Successful for Stream {name}!")

    def sync_wormhole(self, hostname: str):
        logging.debug(f"Wormhole Sync: Syncing With Wormhole Server {hostname}!")

        # Sync information with wormhole server
        resp = requests.post(
            url=f"{hostname}/wormhole/sync",
            json={
                "version": __version__,
                "supported_protocols": list(self.supported_protocols.keys())
            }
        )

        if resp.status_code != 200 and resp.status_code != 400:
            raise Exception(f"Failed To Sync With Wormhole Server! Error: [{resp.status_code}] {resp.text}")

        # Get the response
        resp_json = resp.json()
        if not resp_json:
            raise Exception("Failed To Sync With Wormhole Server! Invalid Json Response!")
        logging.debug(f"Wormhole Sync: Server Responded With: {resp_json}!")

        # Verify that all json fields are present
        # Yes, match exists, BUT this program is supposed to support python3.6 onwards
        if not all(key in resp_json for key in ["ready", "message", "version", "supported_protocols", "managed_streams"]):
            raise Exception("Failed To Sync With Wormhole Server! Missing Json Fields!")

        # Check if errored!
        if resp.status_code == 400:
            raise Exception(f"Failed To Sync With Wormhole Server! {resp_json.get('message', False)}")

        # NOTE: These should be checked by the server already, but just to double check, run again!
        # Check if server is ready
        if not resp_json.get("ready", False):
            raise Exception(f"Failed To Sync With Wormhole Server! Unknown Error: {resp_json}")

        # Check if version matches
        if resp_json.get("version", "N/A") != __version__:
            raise Exception(f"Failed To Sync With Wormhole Server! Version Mismatch! Server: {__version__} Client: {resp_json.get('version', 'N/A')}")

        # Check if server supports all protocols
        if not any([p in resp_json.get("supported_protocols", []) for p in self.supported_protocols.keys()]):
            raise Exception(f"Failed To Sync With Wormhole Server! Server Does Not Have Any Supported Protocols!")

        # Return with server information
        logging.debug(f"Wormhole Sync: Server Sync Finished!")
        return resp_json.get("managed_streams", [])

    def sync_stream(self, hostname: str, name: str):
        logging.debug(f"Wormhole Sync: Syncing With Stream {name} On Wormhole Server {hostname}!")

        # Sync information with wormhole server
        resp = requests.get(
            url=f"{hostname}/wormhole/stream/{name}/sync"
        )

        if resp.status_code != 200:
            raise Exception(f"Failed To Sync With Wormhole Stream! Error: [{resp.status_code}] {resp.text}")

        # Get the response
        resp_json = resp.json()
        if not resp_json:
            raise Exception("Failed To Sync With Wormhole Stream! Invalid Json Response!")
        logging.debug(f"Wormhole Sync: Server Responded With: {resp_json}!")

        # Verify that all json fields are present
        # Yes, match exists, BUT this program is supposed to support python3.6 onwards
        if not all(key in resp_json for key in ["version", "stream_name", "supported_protocols", "stream_info"]):
            raise Exception("Failed To Sync With Wormhole Stream! Missing Json Fields!")

        # Verify Stream Response
        stream_info = resp_json.get("stream_info")
        if not all(key in stream_info for key in ["width", "height", "pixel_size", "max_fps"]):
            raise Exception("Failed To Sync With Wormhole Stream! Stream Info is Missing Json Fields!")

        logging.debug(f"Wormhole Sync: Stream Sync Finished!")
        return resp_json.get("supported_protocols"), stream_info.get("width", 0), stream_info.get("height", 0), stream_info.get("pixel_size", 0), stream_info.get("max_fps", 0)

    #
    # --- Helper Streaming Functions ---
    #

    def create_stream(self, streamer: Type[AbstractStreamer], *args, **kwargs):
        # Initialize Streamer
        streamer_obj = streamer(self.controller, *args, **kwargs)
        # Add Streamer to Wormhole
        self.routes[streamer_obj.route] = streamer_obj

    def get_video(self, name: str = "default"):
        # Check if advanced features are enabled
        if not self.advanced_features:
            raise Exception("Managed Streams Are Only Enabled If Advanced Features Are Enabled!")

        name = name.lower()
        if name not in self.managed_streams:
            raise Exception(f"Stream {name} Was Not Found!")

        return self.managed_streams[name][0]

    def join(self):
        # Joins the wormhole thread so that the app does not exit
        self.wormhole_thread.join()

    def generate_debug_html(self):
        output = ""
        output += "<h2>Debug Information</h2>"
        output += f"<h3>Advanced Features Enabled: {self.advanced_features}</h3>"
        output += f"<h3>Supported Protocols:</h3>"
        for proto, (streamer, viewer) in self.supported_protocols.items():
            output += f"<p>[{proto}] Streamer: {escape(streamer)} | Viewer: {escape(viewer)}<p>"
        output += f"<h3>Enabled Routes:</h3>"
        for route, streamer in self.routes.items():
            output += f"<p>{route} | Streamer: {escape(streamer)}</p>"
        output += f"<h3>Managed Streams:</h3>"
        for name, (video, protocols) in self.managed_streams.items():
            output += f"<p>{name} | Video Object: {escape(video)} | Supported Protocols: {protocols}</p>"
        output += f"<h3>Tunning Threads:</h3>"
        for thread_id, thread in enumerate(threading.enumerate()):
            output += f"<p>{thread_id} | {escape(thread)}</p>"
        return output


class WormholeClient():
    """
    TODO: Lightweight Wormhole Class for Just Reading Streams
    """
    pass



from wormhole.utils import FrameController
from wormhole.video import AbstractVideo

from pathlib import Path
from threading import Thread
from typing import Optional
import cv2
import logging


def write_video(
    video: AbstractVideo,
    filename: str,
    encoding: Optional[str | tuple | int] = None,
    width: Optional[int] = None,
    height: Optional[int] = None,
    max_fps: Optional[float] = None,
    print_fps: bool = False
):
    width = width or video.width
    height = height or video.height
    max_fps = max_fps or video.max_fps
    frame_controller = FrameController(max_fps, print_fps=print_fps)

    # Get File Encoding Format
    if isinstance(encoding, int):
        fourcc = encoding
    elif isinstance(encoding, tuple) or isinstance(encoding, str):
        fourcc = cv2.VideoWriter_fourcc(*encoding)
    else:
        # Autodetect File Encoding
        suffix = Path(filename).suffix
        if suffix == ".mp4":
            # NOTE: h264 is not supported everywhere
            # fourcc = cv2.VideoWriter_fourcc(*"h264")
            fourcc = cv2.VideoWriter_fourcc(*"MP4V")
        elif suffix == ".avi":
            fourcc = cv2.VideoWriter_fourcc(*"XVID")
        elif suffix == ".webm":
            # NOTE: VP9 is wayyy to buggy for handling unstable video
            # fourcc = cv2.VideoWriter_fourcc(*"VP90")
            fourcc = cv2.VideoWriter_fourcc(*"VP80")
        else:
            # TODO: Potentially support more codecs and formats in the future?
            logging.warning("Unknown File Encoding Format. Using OepnCV Default.")
            fourcc = -1

    # Create File Writer Object
    video_writer = cv2.VideoWriter(filename, fourcc, max_fps, (width, height))

    # Sanity Check
    if not video_writer.isOpened():
        raise Exception("Video Writer Failed to Initialize!")

    while True:
        # Sanity Check
        if not video_writer.isOpened():
            raise Exception("Video Writer Suddenly Failed!")

        # Get the next frame
        frame = video.get_frame()

        # Resize frame or else the video will break
        frame = cv2.resize(frame, (width, height))

        # Write the frame to file
        video_writer.write(frame)
        frame_controller.next_frame()

    # This never gets called...
    # video_writer.release()


# TODO: This might cause data loss if the program is closed during video write. Ill look into it later.
def threaded_video_writer(*args, **kwargs):
    # Create a new video writer thread
    video_writer_thread = Thread(target=write_video, args=args, kwargs=kwargs)
    video_writer_thread.daemon = True
    video_writer_thread.start()
    return video_writer_thread



from wormhole.utils import FrameController
from wormhole.video import AbstractVideo

import cv2
from threading import Thread
from typing import Any, Optional


class CameraVideo(AbstractVideo):
    """
    Creates a video object from a camera
    """

    def __init__(
        self,
        cam_id: int,
        max_fps: Optional[float] = 30,
        width: Optional[int] = None,
        height: Optional[int] = None,
        cv2_config: Optional[list[tuple[Any, Any]]] = None,
        **kwargs  # Any Additional Arguments for AbstractVideo
    ):
        # Basic Video Properties
        self.cam_id: int = cam_id

        # Open Camera
        self.cap = cv2.VideoCapture(self.cam_id)

        # Check if camera open was successful
        if self.cap is None or not self.cap.isOpened():
            raise ValueError(f"Camera Not Opened! Unable to open camera {self.cam_id}")

        # Set CV2 Settings
        if cv2_config:
            for key, value in cv2_config:
                self.cap.set(key, value)
        # Set width, height, or fps if they are not set by default
        max_fps = max_fps or int(self.cap.get(cv2.CAP_PROP_FPS))
        width = width or int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = height or int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

        # Initialize Video Object
        super().__init__(width, height, max_fps, **kwargs)

        # Check if Camera Opened Successfully
        if not self.cap.isOpened():
            raise ValueError("Camera Not Opened! An Error Probably Occurred.")
        # Set up Frame Controller
        self.frame_controller = FrameController(self.max_fps, print_fps=self.print_fps)

        # Start Video Thread
        self.video_thread = Thread(target=self.video_loop, daemon=True)
        self.video_thread.start()

    def video_loop(self):
        # Start Video Loop
        while True:
            try:
                # Read Frame
                _, frame = self.cap.read()

                # If sizes does not match, resize frame
                frame_width = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
                frame_height = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                if frame_width != self.width or frame_height != self.height:
                    frame = cv2.resize(frame, (self.width, self.height))

                # Set Frame
                self.set_frame(frame)
                self.frame_controller.next_frame()
            except Exception as e:
                self.handle_render_error(e, message="Error While Capturing Camera Frame!")



from wormhole.video import AbstractVideo


class AbstractViewer(AbstractVideo):
    """
    General Abstract Viewer Class for Wormhole. 
    """

    def __init__(
        self,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)



from wormhole.streamer import AbstractStreamer
from wormhole.utils import FrameController

import cv2
import logging
import time
import traceback
from flask.wrappers import Response
from typing import Optional, Any


class MJPEGStreamer(AbstractStreamer):
    """
    Streamer for the Motion JPEG video protocol
    """

    def __init__(
        self,
        *args,
        boundary: str = "WORMHOLE",
        imencode_config: Optional[list[Any]] = None,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.boundary = boundary
        self.imencode_config = imencode_config

        # Create Video Feed Handler for Flask
        def video_feed():
            # Render and Send Frames for each client
            def generate_next_frame():
                frame_controller = FrameController(self.max_fps, print_fps=self.print_fps)
                while True:
                    try:
                        _, jpg = cv2.imencode(
                            ".jpg",
                            self.video.get_frame(),
                            self.imencode_config or []  # Use Empty Config if imencode is not available
                        )
                        yield (b"--" + boundary.encode("ascii") + b"\r\nContent-Type: image/jpeg\r\n\r\n" + jpg.tobytes() + b"\r\n")
                        frame_controller.next_frame()
                    except Exception as e:
                        # Print Error To User
                        logging.error(f"Error While Generating JPEG for Stream! {e}")
                        traceback.print_exc()
                        time.sleep(1)

                        # Reset FPS Statistics in case the video works again
                        frame_controller.reset_fps_stats()

            return Response(
                generate_next_frame(),
                mimetype=f"multipart/x-mixed-replace; boundary={boundary}",
            )

        # Add the video feed route to the network controller
        self.controller.add_route(self.route, video_feed, strict_url=self.strict_url)



import logging
from flask import Flask
from flask_socketio import SocketIO
from flask_cors import CORS
from typing import Any, Callable, Optional
from uuid import uuid4


class AbstractController():
    """
    Base Class for All Network Controllers
    """

    def __init__(self, *args, **kwargs):
        raise NotImplementedError()

    def get_app(self):
        raise NotImplementedError()

    def add_route(self, route: str, handler: Callable, *args, **kwargs):
        raise NotImplementedError()

    def add_message_handler(self, message: str, handler: Callable, *args, **kwargs):
        raise NotImplementedError()

    def start_server(self, *args, **kwargs):
        raise NotImplementedError()

    def stop_server(self, *args, **kwargs):
        raise NotImplementedError()


class FlaskController(AbstractController):
    """
    Flask Network Controller with a SocketIO Backend
    """

    def __init__(
        self,
        wormhole,
        cors: bool = True,
        host: str = "0.0.0.0",
        port: int = 8000,
        flask_config: dict[str, Any] = {},
        debug: bool = False,
        *args,
        **kwargs
    ):
        # Check if script is being run from an interactive shell
        # If it is, return warning / error message
        import __main__ as main
        if not hasattr(main, '__file__'):
            logging.warning("Wormhole does not work in an interactive shell!")
            logging.warning("Please run your script from a file.")
            logging.warning("This is a known issue that is WIP.")

        # Create The Flask Server
        self.wormhole = wormhole
        self.app = Flask(__name__)
        self.host = host
        self.port = port
        self.debug = debug
        self.server_args = args
        self.server_kwargs = kwargs
        logging.debug(f"Initializing Flask Server on {host}:{port}")
        logging.debug(f"Flask Arguments: {args}, {kwargs}")

        # Set Flask Config Vars
        for k, v in flask_config.items():
            self.app.config[k] = v

        # Setup socketio
        self.socketio = SocketIO(self.app, logger=self.debug, engineio_logger=self.debug, cors_allowed_origins="*")

        # Enable CORS on the server
        if cors:
            CORS(self.app)

        # Monkey Patch _is_setup_finished so that dynamic route additions are allowed
        self.app._is_setup_finished = lambda: False  # type: ignore

    def get_app(self):
        return self.app

    def add_route(self, route: str, handler: Callable, *args, strict_url: bool = True, **kwargs):
        logging.debug(f"Adding Route {route} with handler {handler}")
        # Validate Route
        if not route.startswith('/'):
            raise ValueError("Route must start with '/'")
        if route in self.wormhole.routes:
            raise ValueError(f"Route {route} already exists")
        if self.wormhole.advanced_features and strict_url and route.startswith('/wormhole'):
            logging.warning("The 'wormhole' keyword in the route is reserved. Using it may cause issues.")

        # Hot-add handler for route
        self.app.add_url_rule(route, endpoint=str(uuid4()), view_func=handler, *args, **kwargs)

    def add_message_handler(self, message: str, handler: Callable, namespace: Optional[str] = None, *args, strict_url: bool = True, **kwargs):
        logging.debug(f"Adding SocketIO Message Handler for Message {message} on namespace {namespace} with handler {handler}")
        # Validate Namespace
        if namespace:
            if not namespace.startswith('/'):
                raise ValueError("Route must start with '/'")
            if namespace in self.wormhole.routes:
                raise ValueError(f"Namespace {namespace} already exists")
            if self.wormhole.advanced_features and strict_url and namespace.startswith('/wormhole'):
                logging.warning("The 'wormhole' keyword in the namespace is reserved. Using it may cause issues.")

        # Hot-add handler for socketio message
        self.socketio.on(message, namespace=namespace, *args, **kwargs)(handler)
        # Warning: a little cursed :/

    def start_server(self, *args, **kwargs):
        logging.info(f"Starting Flask Server on {self.host}:{self.port}")
        self.socketio.run(
            self.app,
            log_output=self.debug,
            host=self.host,
            port=self.port,
            debug=self.debug,
            use_reloader=False,
            *[*args, *self.server_args],  # Merge *args and *self.server_args
            **{**kwargs, **self.server_kwargs}  # Merge **kwargs and **self.server_kwargs
        )



from wormhole.video import AbstractVideo

import cv2
import numpy as np
from threading import Thread


class SoftCopy(AbstractVideo):
    """
    Creates a soft copy of another video stream. Uses Frame Subscribers to achieve this effect.
    """

    def __init__(
        self,
        original: AbstractVideo,
        **kwargs  # Any Additional Arguments for AbstractVideo
    ):
        # Initialize Video Object with the original parameters
        super().__init__(original.width, original.height, original.max_fps, **kwargs)

        # Create a subscriber for the other video stream
        def video_update_subscriber(video):
            self.set_frame(np.copy(video.get_frame()))
            self.frame_controller.next_frame()
        original.add_frame_subscriber(video_update_subscriber)


class HardCopy(AbstractVideo):
    """
    Creates a hard copy of another video stream. Uses its own frame controller to achieve this effect
    """

    def __init__(
        self,
        original: AbstractVideo,
        width: int,
        height: int,
        max_fps: float = 30,
        **kwargs  # Any Additional Arguments for AbstractVideo
    ):
        # Initialize Video Object with the NEW parameters
        super().__init__(width, height, max_fps, **kwargs)
        self.original = original

        # Start Video Thread
        self.video_thread = Thread(target=self.video_loop, daemon=True)
        self.video_thread.start()

    def video_loop(self):
        # Start Video Loop
        while True:
            try:
                # Get the new video data
                new_frame = np.copy(self.original.get_frame())

                # If sizes does not match, resize frame
                if self.original.width != self.width or self.original.height != self.height:
                    new_frame = cv2.resize(new_frame, (self.width, self.height))

                # Set Frame Size
                self.set_frame(new_frame)
                self.frame_controller.next_frame()
            except Exception as e:
                self.handle_render_error(e, message="Error While Reading Video Copy!")



from wormhole.viewer import AbstractViewer

import math
import socketio
from typing import Callable, Optional
from urllib.parse import urlparse


class SocketIOViewerBase(AbstractViewer):
    """
    Base Class for Everything SocketIO Viewer
    """

    def __init__(
        self,
        data_processor: Callable,
        url: str,
        width: int,
        height: int,
        max_fps: float = math.inf,
        socketio_args: Optional[dict] = None,
        **kwargs
    ):
        # Save basic variables about stream
        parsed_url = urlparse(url)
        self.hostname = f"{parsed_url.scheme}://{parsed_url.netloc}"
        self.namespace = parsed_url.path

        # Save Raw Data Processing Function
        self.data_processor = data_processor

        # Setup SocketIO Client
        self.sio_client = socketio.Client(**socketio_args if socketio_args else {})

        # Initiate Video Parent Object
        super().__init__(width, height, max_fps=max_fps, **kwargs)

        # Create SocketIO Handler for when raw images stream in
        # Proxying the function with a lambda so that the self context is also passed in
        self.sio_client.on("frame", self.data_processor, namespace=self.namespace)

        # Connect To Server
        self.sio_client.connect(self.hostname, namespaces=[self.namespace])



from wormhole.utils import FrameController
from wormhole.video import AbstractVideo

import cv2
from typing import Optional


def render_video(
    video: AbstractVideo,
    width: Optional[int] = None,
    height: Optional[int] = None,
    max_fps: Optional[float] = None,
    print_fps: bool = False,
    window_name="Video Preview"
):
    # Basic Video Information
    width = width or video.width
    height = height or video.height
    max_fps = max_fps or video.max_fps
    frame_controller = FrameController(max_fps, print_fps=print_fps)

    # Hot loop for video rendering
    while True:
        frame = video.get_frame()
        cv2.imshow(window_name, cv2.resize(frame, (width, height)))
        # Check if q is sent to exit video
        if cv2.waitKey(1) == ord('q'):
            break
        frame_controller.next_frame()

    # Safely clean all windows
    cv2.destroyAllWindows()



from wormhole.utils import FrameController
from wormhole.streamer import AbstractStreamer

import logging
import time
import traceback
from flask_socketio import emit, join_room
from threading import Thread
from typing import Any, Callable


class SocketIOStreamerBase(AbstractStreamer):
    """
    Base Class for Everything SocketIO Streamer
    """

    def __init__(
        self,
        frame_publisher_hotloop: Callable,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        # Main loop to run for video streams
        self.frame_publisher_hotloop = frame_publisher_hotloop

        # Control variables to save on execution when no clients are connected
        self.thread_running = False
        self.connected_clients = 0

        # Setup Background Video Thread
        self.video_streamer_thread = Thread(target=self.video_streamer)

        # Add Connect Handler
        def on_connect():
            self.connected_clients += 1
            # Start the video streamer if it is not already running
            if self.thread_running == False:
                self.thread_running = True
                self.video_streamer_thread = Thread(target=self.video_streamer)
                self.video_streamer_thread.start()

            # Tell client to join the video room
            join_room("video_feed")
        self.controller.add_message_handler("connect", on_connect, namespace=self.route, strict_url=self.strict_url)

        # Add Disconnect Handler
        def on_disconnect():
            self.connected_clients -= 1

            # If no clients are connected anymore, stop server
            if self.connected_clients == 0:
                self.thread_running = False
        self.controller.add_message_handler("disconnect", on_disconnect, namespace=self.route, strict_url=self.strict_url)

    # Streamer simple  for raw video
    def video_streamer(self):
        # Get the flask context
        with self.controller.get_app().app_context():
            frame_controller = FrameController(self.max_fps, print_fps=self.print_fps)
            while True:
                if not self.thread_running:
                    return  # Kill the thread if no clients are connected

                # Run Stream Publisher Function
                # This should process any video encoding work and publishing logic
                try:
                    self.frame_publisher_hotloop()
                except Exception as e:
                    # Print Error To User
                    logging.error(f"Error While Generating Frame for SocketIO Stream! {e}")
                    traceback.print_exc()
                    time.sleep(1)

                    # Reset FPS Statistics in case the video works again
                    frame_controller.reset_fps_stats()

                frame_controller.next_frame()

    # Helper function to emit socket information
    def send_data(self, data: Any):
        emit("frame", data, room="video_feed", namespace=self.route, broadcast=True)

        time.sleep(0)  # Force A Network Buffer Flush
        # Technically this should be a gevent.sleep(0) but time.sleep is monkey patched so this should be alright.

        return True



from wormhole.version import __version__

# Setup Gevent Monkey Patching
try:
    from gevent import monkey
    monkey.patch_all()
except ModuleNotFoundError:
    print("Gevent is not installed. Ignoring Monkey Patch.")

from wormhole.core import *



from wormhole.utils import FrameController
from wormhole.video import AbstractVideo

import numpy as np
from threading import Thread
from typing import Callable, Optional


class CustomVideo(AbstractVideo):
    """
    Creates a video object from a custom video stream.
    Takes in a video_generator function that returns a new frame every time it is called.
    """

    def __init__(
        self,
        width: int,
        height: int,
        max_fps: float,
        frame_generator: Optional[Callable] = None,
        **kwargs  # Any Additional Arguments for AbstractVideo
    ):
        # Basic Video Properties
        self.frame_generator: Optional[Callable] = frame_generator

        # Initialize Video Object
        super().__init__(width, height, max_fps, **kwargs)

        # Set up Frame Controller
        self.frame_controller = FrameController(self.max_fps, print_fps=self.print_fps)

        # Start Video Thread
        # If frame generator is defined, start the video thread.
        # ELSE, the user probably wants to set frames manually by calling set_frame(), so dont start thread
        if self.frame_generator:
            self.video_generator_thread = Thread(target=self.video_loop, daemon=True)
            self.video_generator_thread.start()
        else:
            pass

    def video_loop(self):
        # Start Video Loop
        while True:
            try:
                if self.frame_generator:
                    # Get a new frame from the frame generator
                    frame = self.frame_generator(self)
                    # The frame generator might've already set the frame using set_frame(),
                    # so only run set_frame again if the return type is of ndarray
                    if isinstance(frame, np.ndarray):
                        self.set_frame(frame)
                    self.frame_controller.next_frame()
                else:
                    raise Exception("Frame generator not defined during video loop! This should not happen normally")
            except Exception as e:
                self.handle_render_error(e, message="Frame Generator Encountered An Error!")



from wormhole.viewer import AbstractViewer

import cv2
import math
import logging
import numpy as np
import urllib.request
from threading import Thread


class MJPEGViewer(AbstractViewer):
    """
    Viewer for the Motion JPEG video protocol
    """

    def __init__(
        self,
        url: str,
        width: int,
        height: int,
        max_fps: float = 30,
        auto_reconnect: bool = True,
        **kwargs  # Any Additional Arguments for AbstractVideo
    ):
        # Save basic variables about stream
        self.url = url
        self.auto_reconnect = auto_reconnect

        # Open Video Link
        self.cap = cv2.VideoCapture(self.url)

        # Initiate Video Parent
        super().__init__(width, height, max_fps, **kwargs)

        # Check if Video File Opened
        if not self.cap.isOpened():
            raise ValueError("Video File Not Opened! An Error Probably Occurred.")

        # Start Video Thread
        self.video_decoder_thread = Thread(target=self.video_decoder, daemon=True)
        self.video_decoder_thread.start()

    def video_decoder(self):
        # Start Video Loop
        while True:
            try:
                # Read Frame
                ret, frame = self.cap.read()
                # Check if Frame is Valid
                if not ret:
                    if self.auto_reconnect:
                        while True:
                            self.cap = cv2.VideoCapture(self.url)
                            if not self.cap.isOpened():
                                self.handle_render_error(Exception("Failed to connect to stream... Retrying in 1 second..."), message="Error While Connecting To Stream")
                            else:
                                logging.info("Successfully reconnected to stream!")
                                break
                    else:
                        self.set_blank_frame()

                # If sizes does not match, resize frame
                frame_width = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
                frame_height = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                if frame_width != self.width or frame_height != self.height:
                    frame = cv2.resize(frame, (self.width, self.height))

                # Set Frame
                self.set_frame(frame)
                self.frame_controller.next_frame()
            except Exception as e:
                self.handle_render_error(e, message="Error While Processing/Opening Motion JPEG stream!")


class BufferedMJPEGViewer(AbstractViewer):
    """
    Alternative implementation of the MJPEGViewer class
    """

    def __init__(
        self,
        url: str,
        width: int,
        height: int,
        max_fps: float = math.inf,
        read_buffer_size: int = 1024,
        auto_reconnect: bool = True,
        **kwargs  # Any Additional Arguments for AbstractVideo
    ):
        # Save basic variables about stream
        self.url = url
        self.auto_reconnect = auto_reconnect

        # Save advanced variables about stream
        self.read_buffer_size = read_buffer_size

        # Start the video decoder in another thread
        self.video_decoder_thread = Thread(target=self.video_decoder, daemon=True)
        self.video_decoder_thread.start()

        # Create Object
        super().__init__(width, height, max_fps, **kwargs)

    # Video Decoder Thread
    def video_decoder(self):
        while True:
            try:
                with urllib.request.urlopen(self.url) as stream:
                    # Read jpeg image from stream
                    inBytes = bytes()

                    # TODO: Find a way to make this faster
                    while True:
                        # Find start and end of image file
                        inBytes += stream.read(self.read_buffer_size)
                        a = inBytes.find(b'\xff\xd8')
                        b = inBytes.find(b'\xff\xd9')
                        if a != -1 and b != -1:
                            # Extract image from stream
                            jpg = inBytes[a:b + 2]
                            inBytes = inBytes[b + 2:]

                            # Decode Image
                            np_image = np.fromstring(jpg, dtype=np.uint8)  # type: ignore
                            new_frame = cv2.imdecode(np_image, cv2.IMREAD_COLOR)

                            # If sizes does not match, resize frame
                            frame_height, frame_width, _ = new_frame.shape
                            if frame_width != self.width or frame_height != self.height:
                                new_frame = cv2.resize(new_frame, (self.width, self.height))

                            # Set the frame information
                            self.set_frame(new_frame)
                            self.frame_controller.next_frame()

            # Catch any errors that may occur in the video decoder
            except Exception as e:
                self.handle_render_error(e, message="Error While Processing/Opening Motion JPEG stream!")



from wormhole.utils import FrameController
from wormhole.video import AbstractVideo

import cv2
import math
from threading import Thread
from typing import Optional


class ImageVideo(AbstractVideo):
    """
    Creates a video object from an image file.
    """

    def __init__(
        self,
        filename: str,
        max_fps: Optional[float] = math.inf,
        width: Optional[int] = None,
        height: Optional[int] = None,
        **kwargs  # Any Additional Arguments for AbstractVideo
    ):
        # Basic Video Properties
        self.filename: str = filename

        # Open Image File
        self.image = cv2.imread(self.filename)

        # Check if the image opens properly
        if self.image is None:
            raise ValueError("Image File Not Opened! An Error Probably Occurred.")

        # Get info from image
        image_height, image_width = self.image.shape[:2]

        # Set width, height or fps if they are not set by default
        width = width or int(image_width)
        height = height or int(image_height)
        max_fps = max_fps or math.inf

        # Initialize Video Object
        super().__init__(width, height, max_fps, **kwargs)

        # Render the first instance of the image
        self.render()

        # Set up Frame Controller
        self.frame_controller = FrameController(self.max_fps, print_fps=self.print_fps)

        # If the user defined an FPS to use, start the image rendering thread
        if self.max_fps != math.inf:
            # Start Video Thread
            self.image_rendering_thread = Thread(target=self.video_loop, daemon=True)
            self.image_rendering_thread.start()

    def render(self, reopen_file: bool = False):
        # If user requests to reopen image file, do so
        if reopen_file:
            # Open Image File
            self.image = cv2.imread(self.filename)

            # Check if the image opens properly
            if self.image is None:
                raise ValueError("Image File Not Opened! An Error Probably Occurred.")

        # Get the new image
        new_frame = self.image.copy()

        # If sizes does not match, resize image
        frame_height, frame_width, _ = self.image.shape
        if frame_width != self.width or frame_height != self.height:
            new_frame = cv2.resize(new_frame, (self.width, self.height))

        # Set this image as the new frame
        self.set_frame(new_frame)

    def video_loop(self):
        # Start Video Loop
        while True:
            try:
                # Run render function to generate next frame
                self.render(reopen_file=True)
                self.frame_controller.next_frame()
            except Exception as e:
                self.handle_render_error(e, message="Error While Rendering Image")



# Quick python file to define the version

# If gitpython is installed and the repo is directly from git, use the git version
try:
    import git
    from pathlib import Path
    repo = git.Repo(Path(__file__).parent.parent)  # type: ignore
    __version__ = f"development-{repo.head.object.hexsha[:7]}"  # type: ignore
except Exception as e:
    __major__ = 1
    __minor__ = 0
    __patch__ = 0
    __version__ = f"{__major__}.{__minor__}.{__patch__}"



import setuptools

# Setup the markdown
readme_markdown = ""
ignore = False
for l in open("README.md"):
    if l.startswith("[HTMLElementBegin]"):
        ignore = True
        continue
    elif l.startswith("[HTMLElementEnd]"):
        ignore = False
        continue
    else:
        readme_markdown += l if not ignore else ""
        
# Setup the requirements
requirements = []
for l in open("requirements.txt"):
    requirements.append(l.strip())

setuptools.setup(
    name='wormhole-streaming',
    version="0.0.1.post1",
    description='Simple and Hackable Realtime Video Streaming Engine',
    author='Edward Li',
    url='https://github.com/RadioactiveHydra/Wormhole/',
    python_requires='>=3.9',
    packages=['wormhole', 'wormhole.streamer', 'wormhole.video', 'wormhole.viewer', 'wormhole.assets'],
    package_data={'wormhole.assets':['*']},
    install_requires=requirements,
    long_description=readme_markdown,   
    long_description_content_type='text/markdown',
    project_urls={
        "Source Code": "https://github.com/RadioactiveHydra/Wormhole",
        "Bug Reports": "https://github.com/RadioactiveHydra/Wormhole/issues",
        "Demo": "https://demo.wormhole.hydranet.dev/",
        "Discord": "https://discord.gg/9QF2bPc",
    }
)



import cv2
import math
import numpy as np
import time
from pathlib import Path
from typing import Union


#
# --- Helper functions to render colors, text, and images ---
#


def blank_frame_color(width: int, height: int, color: tuple, pixel_size: int = 3):
    """
    Set the current frame to a blank frame with a given color
    """

    new_frame = np.zeros((height, width, pixel_size), np.uint8)
    new_frame[:, :, :] = color
    return new_frame


def blank_frame_color_rgb(width: int, height: int, r: int, g: int, b: int, pixel_size: int = 3):
    """
    Set the current frame to a blank frame with a given color
    """

    new_frame = np.zeros((height, width, pixel_size), np.uint8)
    new_frame[:, :, :] = (r, g, b)
    return new_frame


def draw_text(
    frame: np.ndarray,
    text: str,
    position: tuple[int, int],
    font_family: int = cv2.FONT_HERSHEY_SIMPLEX,
    font_size: float = 1,
    font_color: tuple = (255, 255, 255),
    font_stroke: int = 2
):
    """
    Draws text at given location
    """

    return cv2.putText(
        frame,
        text,
        position,
        font_family,
        font_size,
        font_color,
        font_stroke
    )


def draw_multiline_text(
    frame: np.ndarray,
    width: int,
    height: int,
    position: tuple[int, int],
    text: Union[str, list[str]]
):
    """
    Helper function to draw multiple lines of text at once. Automatically determines ideal rendering size
    """

    # Parse the text input
    if type(text) is str:
        text = [text]

    # Disable this feature for now.
    '''
    # Determine Information On Rendered Text
    CHAR_WIDTH, CHAR_HEIGHT = 30, 30
    FONT_SIZE = 1
    FONT_STROKE = 2
    num_lines = len(text)
    max_width = max(len(l) for l in text)

    # Check if the text will fit on the screen. If not, zoom out
    if max_width * CHAR_WIDTH < width and num_lines * CHAR_HEIGHT < height:
        pass
    else:
        # Does not fit. Change text size
        CHAR_WIDTH, CHAR_HEIGHT = 15, 15
        FONT_SIZE = 0.5
        FONT_STROKE = 1
    '''

    # Break down position
    x_offset, y_offset = position

    # Always use zoomed out mode
    CHAR_HEIGHT = 18
    FONT_SIZE = 0.5
    FONT_STROKE = 1

    # Place the text on the frame
    for i, line in enumerate(text):
        frame = draw_text(frame, line, (x_offset + 10, y_offset + (i + 1) * CHAR_HEIGHT), font_size=FONT_SIZE, font_stroke=FONT_STROKE)

    return frame


def draw_overlay(frame, overlay_image, position: tuple[int, int], overlay_size: tuple[int, int]):
    """
    Draws a solid image ontop of another imge
    """

    # Extract the positions
    pos_x, pos_y = position

    # Extract the size
    overlay_width, overlay_height = overlay_size

    # If sizes does not match, resize frame
    frame_height, frame_width, _ = overlay_image.shape
    if (frame_width, frame_height) != overlay_size:
        overlay_image = cv2.resize(overlay_image, overlay_size)

    frame[pos_y:pos_y + overlay_height, pos_x:pos_x + overlay_width] = overlay_image

    return frame


def draw_transparent_overlay(frame, overlay_image, position: tuple[int, int], overlay_size: tuple[int, int]):
    """
    Overlay an image with a transparency layer onto a frame
    Code adapted from: https://gist.github.com/clungzta/b4bbb3e2aa0490b0cfcbc042184b0b4e
    """
    # Extract the positions
    pos_x, pos_y = position

    # If sizes does not match, resize frame
    frame_height, frame_width, _ = overlay_image.shape
    if (frame_width, frame_height) != overlay_size:
        overlay_image = cv2.resize(overlay_image, overlay_size)

    # Extract the alpha mask of the RGBA image, convert to RGB
    b, g, r, a = cv2.split(overlay_image)
    overlay_color = cv2.merge((b, g, r))

    # Apply some simple filtering to remove edge noise
    mask = cv2.medianBlur(a, 5)

    h, w, _ = overlay_color.shape
    roi = frame[pos_y:pos_y + h, pos_x:pos_x + w]

    # Black-out the area behind the logo in our original ROI
    img1_bg = cv2.bitwise_and(roi.copy(), roi.copy(), mask=cv2.bitwise_not(mask))

    # Mask out the logo from the logo image.
    img2_fg = cv2.bitwise_and(overlay_color, overlay_color, mask=mask)

    # Update the original image with our new ROI
    frame[pos_y:pos_y + h, pos_x:pos_x + w] = cv2.add(img1_bg, img2_fg)

    return frame


def blend_frames(frame1: np.ndarray, frame2: np.ndarray, transparency=0.5):
    """
    Blends two frames together with a transparency value
    """

    # Render the image onto the overlay
    # Draw image over frame with transparency
    cv2.addWeighted(frame2, transparency, frame1, 1 - transparency, 0, frame1)
    return frame1


#
# --- Frame Modifiers to render watermark ---
#


def render_watermark(video):
    """
    Renders a "Powered By Wormhole" watermark at the bottom of the screen
    """

    # Checks if loaded image is already in cache
    wormhole_watermark = globals().get("wormhole_watermark")
    if wormhole_watermark is None:
        # Load the image
        BASE = Path(__file__).parent
        wormhole_watermark = cv2.imread(str(Path(BASE, "assets", "poweredby_white.png")), -1)

        # Cache the image
        globals()["wormhole_watermark"] = wormhole_watermark

    # Dynamically determine the watermark padding distance
    padding = max(video.height // 50, video.width // 50)

    # --- Calculate the watermark position ---

    # Get watermark height and width
    wm_height, wm_width = wormhole_watermark.shape[:2]

    # Get target width and height
    target_width = video.width // 1.5
    target_height = video.height // 6

    # Get ratios of the two targets, and calculate which one is smaller
    width_ratio = target_width / wm_width
    height_ratio = target_height / wm_height
    ratio = min(width_ratio, height_ratio)

    # Get the proper widths and heights
    width = int(wm_width * ratio)
    height = int(wm_height * ratio)

    # Render the watermark onto a copy of the frame
    watermarked_frame = draw_transparent_overlay(video._frame.copy(), wormhole_watermark, (padding, video.height - height - padding), (width, height))

    # Render the watermark with the new width and height
    blend_frames(video._frame, watermarked_frame)

#
# --- Frame Modifiers to render fps and other statistics ---
#


def render_fps(video):
    """
    Render the current fps on the current frame
    """

    # Render the fps
    draw_text(video._frame, f"FPS: {video.frame_controller.average_fps}", (10, 30))


def render_fraps_fps(video):
    """
    Render the current fps in the top right corner like fraps
    """

    # Get the fps
    fps_str = "N/A" if video.frame_controller.average_fps is math.inf else str(int(video.frame_controller.average_fps))  # Fixes a small bug with inf fps
    # Render the fps
    draw_text(video._frame, f"{fps_str}", (video.width - len(fps_str) * 20 - 10, 30), font_color=(0, 0, 0), font_stroke=6)
    draw_text(video._frame, f"{fps_str}", (video.width - len(fps_str) * 20 - 10, 30), font_color=(0, 255, 255))


def render_full_fps(video):
    """
    Renders full FPS statistics on the current frame. Includes: Frame Time, Instantaneous FPS, FPS over X seconds, and Average FPS
    """

    # Render the fps
    draw_multiline_text(video._frame, video.width, video.height, (0, 0), [
        f"Frame Time: {video.frame_controller.frame_time * 1000:.2f} ms",
        f"Instantaneous FPS: {video.frame_controller.instantaneous_fps:.2f}",
        f"FPS over {video.frame_controller.fps_window_delta:.1f} Seconds: {video.frame_controller.fps_window:.2f}",
        f"Average FPS: {video.frame_controller.average_fps:.2f}"
    ])


def render_debug_info(video):
    """
    Renders full debug information about the video stream
    """

    # Render the fps
    from wormhole.version import __version__
    draw_multiline_text(video._frame, video.width, video.height, (0, 0), [
        f"=== [Debug Information] ===",
        f"Wormhole Version: {__version__}",
        f">>> Video Information <<<",
        f"Render Source: {type(video).__name__}",
        f"Height: {video.height} | Width: {video.width} | Pixel Size: {video.pixel_size}",
        f"Maximum Render FPS: {video.max_fps}",

        f">>> Frame Rate Information <<<",
        f"Frame Time: {video.frame_controller.frame_time * 1000:.2f} ms",
        f"Instantaneous FPS: {video.frame_controller.instantaneous_fps:.2f}",
        f"FPS over {video.frame_controller.fps_window_delta:.1f} Seconds: {video.frame_controller.fps_window:.2f}",
        f"Average FPS: {video.frame_controller.average_fps:.2f}",
        f"Frames Rendered: {video.frame_controller.frames_rendered}",

        f">>> Frame Modifiers <<<",
        *[str(modifier) for modifier in video.frame_modifiers],
        f">>> Frame Subscribers <<<",
        *[str(subscriber) for subscriber in video.frame_subscribers],

        f"=== [Debug Information] ===",
    ])


#
# --- Additional Frame Modifiers to modify image ---
#


def grayscale_filter(video):
    """
    Makes video black and white
    """

    cv2.cvtColor(video._frame, cv2.COLOR_BGR2GRAY)
    cv2.cvtColor(video._frame, cv2.COLOR_GRAY2BGR)


def inverse_filter(video):
    """
    Inverts all colors in the video
    """

    cv2.bitwise_not(video._frame)


#
# --- Helper Classes ---
#


class FrameController():
    """
    Helper Class to Control Video Frame Rate
    """

    def __init__(self, fps: float, print_fps: bool = False, sleep_func=None, fps_window_delta: float = 5.0):
        # --- Core Functionality ---
        if fps <= 0.0:
            raise ValueError("FPS must be greater than 0!")

        self.target_fps: float = fps
        self.last_frame: float = time.time()
        self.frame_time: float = 0.0

        # --- Sanity Checks for FPS ---

        # Instantaneous FPS
        self.instantaneous_fps: float = math.inf
        # Average FPS
        self.print_fps: bool = print_fps
        self.start_time: float = time.time()
        self.frames_rendered: int = 0
        self.average_fps: float = math.inf
        # Windowed FPS (FPS over X seconds)
        self.fps_window_delta: float = fps_window_delta
        self.fps_window_start: float = time.time()
        self.fps_window_frames_rendered: int = 0
        self.fps_window: float = math.inf

        # --- Misc Stuff ---

        # Sleep Function is used for if a specific non-blocking sleep function is needed.
        # Gevent Patches already do this, so its not that necessary, but keeping it here for compatibility.
        self.sleep_func = sleep_func or time.sleep

    def get_sleep_time(self):
        cur_time = time.time()
        self.frame_time = cur_time - self.last_frame
        return max((1. / self.target_fps) - self.frame_time, 0.0)

    def next_frame(self):
        # Ensure FPS
        sleep_time = self.get_sleep_time()
        if sleep_time:  # Save on the expensive sleep function is no sleep is needed
            self.sleep_func(sleep_time)
        self.last_frame = time.time()
        # Update FPS Counter
        self.update_fps()

    def update_fps(self):
        # Calculate Instantaneous FPS
        if self.frame_time != 0.0:  # Fix bug where frame is so fast that it divides by zero
            self.instantaneous_fps = 1.0 / self.frame_time
        else:
            self.instantaneous_fps = math.inf
        # Calculate Total Average FPS
        self.frames_rendered += 1
        self.average_fps = self.frames_rendered / (time.time() - self.start_time)
        # Calculate Windowed FPS
        self.fps_window_frames_rendered += 1
        if time.time() - self.fps_window_start > self.fps_window_delta:
            self.fps_window = self.fps_window_frames_rendered / (time.time() - self.fps_window_start)
            self.fps_window_frames_rendered = 0
            self.fps_window_start = time.time()
        # Print FPS
        if self.print_fps:
            print(f"Instantaneous FPS: {self.instantaneous_fps:.2f} Average FPS: {self.average_fps:.2f} FPS over {self.fps_window_delta:.1f} Seconds: {self.fps_window:.2f} Frame Time {self.frame_time * 1000:.2f} ms")

    def reset_fps_stats(self):
        self.__init__(self.target_fps, self.print_fps, sleep_func=self.sleep_func, fps_window_delta=self.fps_window_delta)
